/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*******************************************************************************
* MPIA, NEVEC - MIDI project
*
* "@(#) $Id: visibility.c,v 1.5 2002/11/07 12:50:39 jaffe Exp $"
*
* who        when      what
* ---------  --------  ----------------------------------------------
* jdejong    2002-05-08 imported
*
*/

#define _POSIX_SOURCE_1
#include "vltPort.h"

/* $Id: visibility.c,v 1.5 2002/11/07 12:50:39 jaffe Exp $ */

/*----------------------------------------------------------------------- */
/* Visibility data FITS binary tables routines                            */
/* Partially/fully calibrated data table for optical/IR interferometer    */
/*----------------------------------------------------------------------- */
#include "visibility.h"

/* internal prototypes */
/* get data from table_def specific to visibility table */
void get_visibility_info (visibility_table_def *table);
/* set values in visibility_geom into table keywords */
void visibility_set_geom(visibility_table_def *table);
/* get values in visibility_geom from table keywords */
void visibility_get_geom(visibility_table_def *table);

/* file global variables for table type */
static char *tabletype = "VISIBILITY";


/* visibility_geom functions */
/* create a visibility_geom */
visibility_geom* 
make_visibility_geom(int  naxis[2], char *ctype[2], double crval[2], 
		     float crpix[2], float cd[2][2], char *date_obs)
/*----------------------------------------------------------------------- */
/*   Make an visibility_geom structure                                    */
/*   Inputs:                                                              */
/*      naxis      dimension of two axes of data, weight arrays           */
/*      ctype      labels for the axes                                    */
/*      crval      reference value for each axis                          */
/*      crpix      reference pixels                                       */
/*      cd         cd transformation matrix                               */
/*      date_obs    Date (yyyy-mm-dd) observations began                  */
/*   Returns:                                                             */
/*      An visibility_geom structure pointer ; NULL indicates failure     */
/*----------------------------------------------------------------------- */
{
  visibility_geom *out;
  int i;

  /* create basic structure */
  out = (visibility_geom*)malloc(sizeof(visibility_geom));
  if (out==NULL) return out; /* allocation failure? */
  
  /* zero memory */
  memset (out, 0, sizeof(visibility_geom));

  /* save values */
  for (i=0; i<2; i++) {
    out->naxis[i] = naxis[i];
    out->crval[i] = crval[i];
    out->crpix[i] = crpix[i];
    out->cd[i][0] = cd[i][0];
    out->ctype[i] = (char*)malloc(strlen(ctype[i])+1);
    strcpy (out->ctype[i], ctype[i]);
  }
  
    out->date_obs = (char*)malloc(strlen(date_obs)+1);
    strcpy (out->date_obs, date_obs);
    return out;
} /* end make_visibility_geom */

/* delete a visibility_geom structure */
void kill_visibility_geom (visibility_geom *geom)
/*----------------------------------------------------------------------- */
/*   Delete an visibility_geom structure                                  */
/*   Inputs:                                                              */
/*      geom    pointer to the visibility_geom structure.                 */
/*----------------------------------------------------------------------- */
{
  int i;
  if (geom==NULL) return; /* sanity check */

  for (i=0; i<2; i++) {
    if (geom->ctype[i]!=NULL) free(geom->ctype[i]);
  }
  if (geom->date_obs!=NULL) free(geom->date_obs);

  free (geom); 
} /* end kill_visibility_geom */

/* visibility_row functions */
visibility_row* make_visibility_row(visibility_table_def* table)
/*----------------------------------------------------------------------- */
/*   Make an visibility_row structure                                   */
/*   Inputs:                                                              */
/*      table     pointer to visibility table structure                 */
/*   Returns:                                                             */
/*      An visibility_row structure pointer ; NULL indicates failure    */
/*      Data pointers set to generic structure values                     */
/*----------------------------------------------------------------------- */
{
  visibility_row *out;
  int i, icol, nregion;

  /* create basic structure */
  out = (visibility_row*)malloc(sizeof(visibility_row));
  if (out==NULL) return out; /* allocation failure? */
  
  /* zero memory */
  memset (out, 0, sizeof(visibility_row));
  
  /* Make generic structure */
  out->raw = make_table_row_from_def(table->def);

  /* attach data pointers to those in generic structure*/
  /* When the generic row is read/written the generic array is */
  /* automatically updated */
  icol = table->data_col;  /* which column? */
  out->data = (float*)out->raw->data[icol]->tdata.tvoid; 
  icol = table->weight_col;  /* which column? */
  out->weight = (float*)out->raw->data[icol]->tdata.tvoid; 

  return out; /* done */
} /* end  make_visibility_row */

void kill_visibility_row (visibility_row *row)
/*----------------------------------------------------------------------- */
/*   Delete an visibility_row structure                                   */
/*   Inputs:                                                              */
/*      row    pointer to the visibility_row structure.                   */
/*----------------------------------------------------------------------- */
{
  int i, nregion;
  if (row==NULL) return; /* sanity check */

  free (row->raw); /* generic structure */
  free (row);      /* basic structure */
} /* end  kill_visibility_row */


/* visibility_table functions */
visibility_table_def*
make_visibility_table_def(fitsfile *fptr, visibility_geom *geom,
			  int *ierr)
/*----------------------------------------------------------------------- */
/*   Make a new visibility_table in an open FITS file                     */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      geom  visibility_geom_table describing table                      */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An visibility_table_def pointer; NULL indicates failure           */
/*----------------------------------------------------------------------- */
{
  visibility_table_def *out = NULL;
  int    version = 1; /* table version number */
  int    ikey, icol, i;
  long   nx, ny;

  if (*ierr>0) return NULL;    /* existing error condition? */

  if (fptr==NULL) return NULL; /* sanity check */
  if (geom==NULL) return NULL; /* sanity check */

  /* create output object */
  out = (visibility_table_def*)malloc(sizeof(visibility_table_def));
  if (out==NULL)     {  /* check for allocation failure */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating %s table_def \n", tabletype);
    return out;
  }
  memset (out, 0, sizeof(visibility_table_def)); /* zero fill */

  /* create table_def */
  out->def = make_table_def(VISIBILITY_TABLE, 0, 0);
  if (out->def==NULL) { /* check for failure */
    kill_visibility_table_def(out); /* clean up */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating IMAGING_DATA_TABLE for %s table_def \n", 
	    tabletype);
    return NULL;
  }

  /* attach visibility_geom_table */
  out->geom = geom;

  /* add keywords */
  ikey = table_find_key(out->def, "REVISION");
  if ((ikey>0) && (*ierr==0)) table_put_keyword (out->def, ikey, 
						(void*)&version, ierr);

  /* set keywords from geometry descriptor */
  visibility_set_geom(out);

  /* get info from table descriptor */
  get_visibility_info (out);

  /* set data size */
  icol = out->data_col; 
  /* set data array size and dimension */
  /* get size from Imaging Detector structure */
  nx = geom->naxis[0];
  ny = geom->naxis[1];
  /* data array size */
  out->def->colcount[icol] =  nx*ny;
  /* weight */
  icol = out->weight_col; 
  out->def->colcount[icol] =  nx*ny;

  /* open imaging data table */
  make_table_from_def (fptr, out->def, ierr);

  return out; /* done */
} /* end make_visibility_table */

visibility_table_def* open_visibility_table(fitsfile *fptr,
						int *ierr)
/*----------------------------------------------------------------------- */
/*   Open an existing visibility_table in the active FITS file            */
/*   Fills visibility_geom and saves in visibility_table_def              */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An visibility_table_def pointer; NULL indicates failure           */
/*----------------------------------------------------------------------- */
{
  visibility_table_def *out = NULL;

  if (*ierr>0) return NULL;    /* existing error condition? */
  if (fptr==NULL) return NULL; /* sanity check */

  /* create output object */
  out = (visibility_table_def*)malloc(sizeof(visibility_table_def));
  if (out==NULL)     {  /* check for allocation failure */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating %s table_def \n", tabletype);
    return out;
  }
  memset (out, 0, sizeof(visibility_table_def)); /* zero fill */

  /* position FITS file at  table */
  if (fits_movnam_hdu(fptr, BINARY_TBL, tabletype, 0, ierr)) {
    fprintf(stderr,"ERROR %d finding %s table \n", *ierr, tabletype);
    kill_visibility_table_def(out); /* clean up */
    *ierr = 10000;
    return NULL;
  }

  /* get table descriptor */
  out->def = make_table_def_from_table (fptr);
  if (out->def==NULL) { /* error check */
    fprintf(stderr,"ERROR making table def for %s \n", tabletype);
    kill_visibility_table_def(out); /* clean up */
    *ierr = 10000;
    return NULL;
  }

  /* set geometry descriptor from table keywords */
  visibility_get_geom(out);

  /* copy info from table descriptor */
  get_visibility_info (out);

  return out; /* done */
} /* end open_visibility_table */

int read_visibility_table(fitsfile *fptr, visibility_table_def* table, 
			    visibility_row *row, long row_number,
			    int *ierr)
/*----------------------------------------------------------------------- */
/*   Read the next row from an visibility_table                           */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row_number which row (1-rel) in the table                         */   
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      row       pointer to an extant row structure                      */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  table_row *trow;
  int        i;

  if (*ierr>0) return 1;     /* existing error condition? */
  if (fptr==NULL) return 1;  /* sanity check */
  if (table==NULL) return 1; /* sanity check */
  if (row==NULL) return 1;   /* sanity check */

  /* set row number */
  if (row_number>0) { /* explicit row number */
    row->raw->trow = row_number;
    table->row = row_number;
  } else { /* next row */
    row->raw->trow = ++table->row;
  }

  /* read row */
  table_read_row(fptr, table->def, row->raw, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR  read_visibility_table: reading table %s \n", 
	    tabletype);
    return 1;
  }

  /* copy relevant data from generic table_row */
  /* no need to copy data since pointers are OK */
  trow = row->raw;
  row->time        = trow->data[table->time_col]->tdata.tdouble[0];
  row->timei       = trow->data[table->timei_col]->tdata.tfloat[0];
  row->uvw[0]      = trow->data[table->uvw_col]->tdata.tdouble[0];
  row->uvw[1]      = trow->data[table->uvw_col]->tdata.tdouble[1];
  row->uvw[2]      = trow->data[table->uvw_col]->tdata.tdouble[2];
  row->offset[0]   = trow->data[table->offset_col]->tdata.tfloat[0];
  row->offset[1]   = trow->data[table->offset_col]->tdata.tfloat[1];
  row->rotation    = trow->data[table->rotation_col]->tdata.tfloat[0];
  row->source      = trow->data[table->source_col]->tdata.tshort[0];
  row->tele[0]     = trow->data[table->tele_col]->tdata.tshort[0];
  row->tele[1]     = trow->data[table->tele_col]->tdata.tshort[1];

  return 0; /* OK - done */
} /* end read_visibility_table */

int write_visibility_table(fitsfile *fptr, visibility_table_def* table, 
			     visibility_row *row, long row_number, 
			     int *ierr)
/*----------------------------------------------------------------------- */
/*   Write  the next row to an visibility_table                         */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row       pointer to an extant row structure                      */
/*      row_number which row (1-rel) in the table                         */     
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  table_row *trow;
  int        i;

  if (*ierr>0) return 1;     /* existing error condition? */
  if (fptr==NULL) return 1;  /* sanity check */
  if (table==NULL) return 1; /* sanity check */
  if (row==NULL) return 1;   /* sanity check */

  /* set row number */
  if (row_number>0) { /* explicit row number */
    row->raw->trow = row_number;
    table->row = row_number;
  } else { /* next row */
    row->raw->trow = ++table->row;
  }

  /* copy relevant data to generic table_row */
  /* no need to copy data since pointers are OK */
  trow = row->raw;
  trow->data[table->time_col]->tdata.tdouble[0]    = row->time;
  trow->data[table->timei_col]->tdata.tfloat[0]    = row->timei;
  trow->data[table->uvw_col]->tdata.tdouble[0]     = row->uvw[0];
  trow->data[table->uvw_col]->tdata.tdouble[1]     = row->uvw[1];
  trow->data[table->uvw_col]->tdata.tdouble[2]     = row->uvw[2];
  trow->data[table->offset_col]->tdata.tfloat[0]   = row->offset[0];
  trow->data[table->offset_col]->tdata.tfloat[1]   = row->offset[1];
  trow->data[table->rotation_col]->tdata.tfloat[0] = row->rotation;
  trow->data[table->source_col]->tdata.tshort[0]   = row->source;
  trow->data[table->tele_col]->tdata.tshort[0]     = row->tele[0];
  trow->data[table->tele_col]->tdata.tshort[1]     = row->tele[1];

  /* write row */
  table_write_row(fptr, table->def, row->raw, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR  write_visibility_table: writing table %s \n", 
	    tabletype);
    return 1;
  }

  return 0; /* OK - done */
} /* end write_visibility_table */

int close_visibility_table(fitsfile *fptr, visibility_table_def* table, 
			     int *ierr)
/*----------------------------------------------------------------------- */
/*   Close an open visibility_table, associated structures are destroyed  */
/*   Note: this closes the full FITS file                                 */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  if (*ierr>0) return 1;    /* existing error condition? */
  if (fptr==NULL) return 1; /* sanity check */

  /* do actual file close */
  fits_close_file (fptr, ierr);

  /* delete structures */
  kill_visibility_table_def(table);

  return 0; /* OK - done */
} /* end close_visibility_table */

void kill_visibility_table_def (visibility_table_def* table)
/*----------------------------------------------------------------------- */
/*   Delete an visibility_table structure                                 */
/*   Inputs:                                                              */
/*      table     table definitions structure from open or create         */
/*----------------------------------------------------------------------- */
{
  if (table==NULL) return; /* sanity check */

  /* table definition structure */
  kill_table_def(table->def);
  /* geom table definition structure */
  kill_visibility_geom(table->geom);

  /* free basic structure */
  free (table);
} /* end kill_visibility_table_def */

/*************************private functions ******************************/

void get_visibility_info (visibility_table_def *table)
/*----------------------------------------------------------------------- */
/*   Get data from table_def specific to visibility table                 */
/*   Uses geom member which must already be attached                      */
/*   Inputs:                                                              */
/*      table     table definitions structure                             */
/*----------------------------------------------------------------------- */
{
  if (table==NULL) return;         /* sanity check */
  if (table->geom==NULL) return;   /* sanity check */

  /* lookup column numbers (0-rel) */
  table->time_col      = table_find_col (table->def, "TIME")-1;
  table->timei_col     = table_find_col (table->def, "EXP_TIME")-1;
  table->source_col    = table_find_col (table->def, "SOURCE")-1;
  table->tele_col      = table_find_col (table->def, "TELESCOPE")-1;
  table->uvw_col       = table_find_col (table->def, "UVW")-1;
  table->offset_col    = table_find_col (table->def, "OFFSET")-1;
  table->rotation_col  = table_find_col (table->def, "ROTATION")-1;
  table->data_col      = table_find_col (table->def, "DATA1")-1;
  table->weight_col    = table_find_col (table->def, "WEIGHT1")-1;
} /* end  get_visibility_info */

void visibility_set_geom(visibility_table_def *table)
/*----------------------------------------------------------------------- */
/*   Set table keywords from values in table->geom                        */
/*   Uses geom member which must already be attached                      */
/*   Inputs:                                                              */
/*      table     table definitions structure                             */
/*----------------------------------------------------------------------- */
{
  int i, ierr=0, ikey, datacol;
  char keyword[20];
  visibility_geom *geom;

  if (table==NULL) return;        /* sanity check */
  if (table->geom==NULL) return;  /* sanity check */

  geom = table->geom;
  /* observing date */
  ikey = table_find_key(table->def, "DATE-OBS");
  if ((ikey>0) && (ierr==0)) 
    table_put_keyword (table->def, ikey, (void*)geom->date_obs, &ierr);

  /* geometry given relative to "DATA1" column. */
  datacol = table_find_col (table->def, "DATA1");

  for (i=1; i<=2; i++) {
    /* iNAXISj */
    sprintf (keyword,"%dNAXIS%d", i, datacol);
    ikey = table_find_key(table->def, keyword);
    if ((ikey>0) && (ierr==0)) 
    table_put_keyword (table->def, ikey, (void*)&geom->naxis[i-1], &ierr);

    /* iCTYPj */
    sprintf (keyword,"%dCTYP%d", i, datacol);
    ikey = table_find_key(table->def, keyword);
    if ((ikey>0) && (ierr==0)) 
    table_put_keyword (table->def, ikey, (void*)geom->ctype[i-1], &ierr);

    /* iCRVLj */
    sprintf (keyword,"%dCRVL%d", i, datacol);
    ikey = table_find_key(table->def, keyword);
    if ((ikey>0) && (ierr==0)) 
    table_put_keyword (table->def, ikey, (void*)&geom->crval[i-1], &ierr);

    /* iCRPXj */
    sprintf (keyword,"%dCRPX%d", i, datacol);
    ikey = table_find_key(table->def, keyword);
    if ((ikey>0) && (ierr==0)) 
    table_put_keyword (table->def, ikey, (void*)&geom->crpix[i-1], &ierr);

    /* 1CDij */
    sprintf (keyword,"%d1CD%d", i, datacol);
    ikey = table_find_key(table->def, keyword);
    if ((ikey>0) && (ierr==0)) 
    table_put_keyword (table->def, ikey, (void*)&geom->cd[0][i-1], &ierr);

    /* 2CDij */
    sprintf (keyword,"%d2CD%d", i, datacol);
    ikey = table_find_key(table->def, keyword);
    if ((ikey>0) && (ierr==0)) 
    table_put_keyword (table->def, ikey, (void*)&geom->cd[1][i-1], &ierr);
  }
} /* end  visibility_set_geom */

void visibility_get_geom(visibility_table_def *table)
/*----------------------------------------------------------------------- */
/*   Set values in table->geom from table keywords                        */
/*   Create geom member                                                   */
/*   Inputs:                                                              */
/*      table     table definitions structure                             */
/*----------------------------------------------------------------------- */
{
  double crval[2];  /* reference value for each axis */
  float  crpix[2];  /* reference pixels */
  float  cd[2][2];  /* cd transformation matrix */
  int  naxis[2];    /* dimension of two axes of data, weight arrays */
  char date_obs[11];   /* observing date */
  char *ctype[2];   /* labels for the two axes */
  int ikey, i, ierr=0, datacol;
  char keyword[20];

  if (table==NULL) return;        /* sanity check */
  if (table->geom==NULL) return;  /* sanity check */

  ikey = table_find_key(table->def, "DATE-OBS");
  if ((ikey>0) && (ierr==0)) 
    table_get_keyword (table->def, ikey, (void*)date_obs, &ierr);

  /* geometry given relative to "DATA1" column. */
  datacol = table_find_col (table->def, "DATA1");

  for (i=1; i<=2; i++) {
    /* iNAXISj */
    sprintf (keyword,"%dNAXIS%d", i, datacol);
    ikey = table_find_key(table->def, keyword);
    if (ikey>0) 
      table_get_keyword (table->def, ikey, (void*)&naxis[i-1], &ierr);

    /* iCTYPj */
    sprintf (keyword,"%dCTYP%d", i, datacol);
    ikey = table_find_key(table->def, keyword);
    if (ikey>0) 
      table_get_keyword (table->def, ikey, (void*)ctype[i-1], &ierr);

    /* iCRVLj */
    sprintf (keyword,"%dCRVL%d", i, datacol);
    ikey = table_find_key(table->def, keyword);
    if (ikey>0) 
      table_get_keyword (table->def, ikey, (void*)&crval[i-1], &ierr);

    /* iCRPXj */
    sprintf (keyword,"%dCRPX%d", i, datacol);
    ikey = table_find_key(table->def, keyword);
    if (ikey>0) 
      table_get_keyword (table->def, ikey, (void*)&crpix[i-1], &ierr);

    /* i1CDj */
    sprintf (keyword,"%d1CD%d", i, datacol);
    ikey = table_find_key(table->def, keyword);
    if (ikey>0) 
      table_get_keyword (table->def, ikey, (void*)&cd[0][i-1], &ierr);

    /* i2CDj */
    sprintf (keyword,"%d2CD%d", i, datacol);
    ikey = table_find_key(table->def, keyword);
    if (ikey>0) 
      table_get_keyword (table->def, ikey, (void*)&cd[1][i-1], &ierr);
  }

  table->geom = 
    make_visibility_geom(naxis, ctype, crval, crpix, cd, date_obs);

} /* end visibility_get_geom */
/* */
